# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for generating images performing inference and showing them in browser."""

import os
import sys
import time
import copy
import datetime
import logging

from foglamp.common import logger, iprpc
from foglamp.plugins.common import utils
from foglamp.common.plugin_helpers import HandleMap, PluginHandle

__author__ = "Deepanshu Yadav"
__copyright__ = "Copyright (c) 2021 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


_module_dir = os.path.dirname(__file__)
_handle_map = HandleMap(__name__)

# SETTINGS
_LOGGER = logger.setup(__name__, level=logging.DEBUG)


class MediaHandle(PluginHandle):
    RPC_SERVER_NAME = 'edgeml_server'          # name of RPC server module
    RPC_CONFIG_MEMBERS = ['source', 'fpm', 'source_dir_name', 'camera_id',
                          'rtsp_url', 'inference_choice', 'deployment_name',
                          'rest_url', 'stream_results', 'stream_port',
                          'output_asset', 'destination_dir', 'rotate_after_minutes',
                          'rotate_data_minutes']

    PARAMETERS_THAT_CAUSE_SHUTDOWN = [('source', 'str', 'source'),
                                      ('sourceDirName', 'str', 'source_dir_name'),
                                      ('cameraId', 'int', 'camera_id'),
                                      ('rtspUrl', 'str', 'rtsp_url'),
                                      ('inferenceChoice', 'str', 'inference_choice'),
                                      ('deploymentName', 'str', 'deployment_name'),
                                      ('restUrl', 'str', 'rest_url'),
                                      ('streamResults', 'bool', 'stream_results'),
                                      ('streamPort', 'int', 'stream_port')]

    def __init__(self, config, service_name):
        """[hold the state used to manage the process of capturing and analyzing images]

        Args:
            config ([type]): [The config from the plugin_init API.]
        """
        super().__init__(service_name)

        self.logger.setLevel(logging.DEBUG)

        # creating slots

        self.source = None
        self.fpm = None
        self.source_dir_name = None
        self.camera_id = None
        self.rtsp_url = None
        self.inference_choice = None
        self.deployment_name = None
        self.rest_url = None
        self.stream_results = None
        self.stream_port = None
        self.output_asset = None
        self.destination_dir = None
        self.rotate_after_minutes = None
        self.rotate_data_minutes = None

        self.rpc = None
        self.readings_list = []
        self.reading_iter = None

        self.config_update(config)
        self.preprocess_config()
        self._setup(config)

    def preprocess_config(self):
        """
        Handles data types of config parameters.
        Returns:
            None
        """
        self.source = str(self.source)
        self.fpm = int(self.fpm)
        self.source_dir_name = str(self.source_dir_name)
        self.camera_id = int(self.camera_id)
        self.rtsp_url = str(self.rtsp_url)
        self.inference_choice = str(self.inference_choice)
        self.deployment_name = str(self.deployment_name)
        self.rest_url = str(self.rest_url)
        self.stream_results = bool(self.stream_results)
        self.stream_port = int(self.stream_port)
        self.output_asset = str(self.output_asset)
        self.destination_dir = str(self.destination_dir)
        self.rotate_after_minutes = int(self.rotate_after_minutes)
        self.rotate_data_minutes = int(self.rotate_data_minutes)

    def _setup(self, config):
        """
         Starts the processing at server side.
        Args:
            config (dict): The config of the plugin.

        Returns:

        """
        self.rpc_setup(config, _module_dir)
        # make sure server starts up
        time.sleep(2)
        self.reading_iter = self.get_iterator()

    def get_iterator(self):
        # just in case we have readingPerSec greater than 1.
        while self.readings_list:
            try:
                _reading = self.readings_list.pop(0)
            except IndexError as ie:
                yield None
            yield _reading

    def check_need_to_shutdown(self, new_config):
        """
        Checks whether we need to shutdown.
        Args:
            new_config (dict):

        Returns:
            True or False if shutdown is required.
        """
        self.logger.info("The new config is {}".format(new_config))
        for _name, _type, _snake_case_type in self.PARAMETERS_THAT_CAUSE_SHUTDOWN:
            if _type == 'int':
                _new_value = int(new_config[_name]['value'])
            if _type == 'str':
                _new_value = str(new_config[_name]['value'])
            if _type == 'bool':
                _new_value = bool(new_config[_name]['value'])
            _old_value = getattr(self, _snake_case_type)

            if _old_value != _new_value:
                return True

        return False

    def reconfigure(self, new_config):
        """[Handles reconfigure for the plugin side.]

        Args:
            new_config ([dict]): [The new configuration to be used.]
        """
        # assign handle values from new config

        if self.rpc is not None:
            _need_to_shutdown = self.check_need_to_shutdown(new_config)
            if _need_to_shutdown:
                # we have to call shutdown at server side.
                self.rpc.plugin_shutdown()
                self.rpc_setup(new_config, _module_dir)
            else:
                # calling reconfigure at server side.
                self.config_update(new_config)
                self.preprocess_config()
                members_dict = {}
                for member in self.RPC_CONFIG_MEMBERS:
                    members_dict[member] = getattr(self, member)
                self.rpc.plugin_reconfigure(members_dict)

    def shutdown(self):
        """[calls plugin_shutdown at the server side.]
        """
        if self.rpc is not None:
            self.rpc.plugin_shutdown()
            self.rpc = None

    def poll(self):
        if self.rpc is not None:
            _readings = self.rpc.plugin_poll()
            if len(_readings) > 0:
                for r in _readings:
                    self.readings_list.append(r)
            _reading = next(self.reading_iter, None)
            return _reading

# ### PLUGIN API ### #


_DEFAULT_CONFIG = {
    "plugin": {
        "description": "EdgeML South plugin",
        "type": "string",
        "default": "edgeml",
        "readonly": "true",
    },
    "source": {
        "description": "Source of data being generated",
        "type": "enumeration",
        "options": ["camera", "directory", "stream"],
        "default": "stream",
        "displayName": "Source of Data Generation",
        "order": "2",
    },
    "sourceDirName": {
        "description": "Name of source directory of images",
        "type": "string",
        "default": "Directory For Using Images",
        "displayName": "Directory Name",
        "order": "3",
        "validity": "source == \"directory\"",
    },
    "cameraId": {
        "description": "Id of Camera Used",
        "type": "integer",
        "default": "0",
        "displayName": "Camera ID",
        "order": "4",
        "validity": "source == \"camera\"",
    },
    "rtspUrl": {
        "description": "The URL of RTSP stream",
        "type": "string",
        "default": "rtsp://localhost:8554/clip",
        "displayName": "RTSP URL",
        "order": "5",
        "validity": "source == \"stream\"",
    },
    "fpm": {
        "description": "Frames per minute",
        "type": "integer",
        "default": "1000",
        "displayName": "Frames Per Minute",
        "order": "6",
    },
    "inferenceChoice": {
        "description": "Select the destination to post the images.",
        "type": "enumeration",
        "options": ["k8", "URL"],
        "default": "k8",
        "displayName": "Inference Choice",
        "order": "7",
    },
    "deploymentName": {
        "description": "The deployment name of edgeml cluster.",
        "type": "string",
        "default": "mledge-deployment",
        "displayName": "Deployment Name",
        "order": "8",
        "validity": "inferenceChoice == \"k8\"",
    },
    "restUrl": {
        "description": "The URL to post the images",
        "type": "string",
        "default": "http://localhost:30163/v1/vision/detection",
        "displayName": "The URL to post the images.",
        "order": "9",
        "validity": "inferenceChoice == \"URL\"",
    },
    "streamResults": {
        "description": "Whether to stream results over a browser.",
        "type": "boolean",
        "default": "true",
        "displayName": "Steam Results",
        "order": "10",
    },
    "streamPort": {
        "description": "The port over which we can display detection results in browser.",
        "type": "string",
        "default": "8085",
        "displayName": "Stream Port",
        "order": "11",
        "validity": "streamResults == \"true\"",
    },
    "outputAsset": {
        "description": "The name of asset which contains detected results",
        "type": "string",
        "default": "Detection Results",
        "displayName": "Output Asset",
        "order": "12",

    },
    "destinationDir": {
        "description": "The directory where resultant images will be stored",
        "type": "string",
        "default": "detection",
        "displayName": "Destination Directory",
        "order": "13",
    },
    "rotateAfterMinutes": {
        "description": "Rotate saved images after these many minutes.",
        "type": "string",
        "default": "120",
        "displayName": "Rotate after",
        "order": "14",
    },
    "rotateDataMinutes": {
        "description": "Rotate this much data in minutes",
        "type": "string",
        "default": "10",
        "displayName": "Rotation Data Amount",
        "order": "15",
    },
}


def plugin_info():
    """Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        "name": "South Edge ML",
        "version": "1.9.2",
        "mode": "poll",
        "type": "south",
        "interface": "1.0",
        "config": _DEFAULT_CONFIG,
    }


def plugin_init(config):
    """Initialise the plugin.
    Args:
        config: JSON configuration document for the South plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    _LOGGER.info("PLUGIN INIT, making a new media handle")
    _media_handle = MediaHandle(config, "south-edgeml")
    _h = _handle_map.new_handle(_media_handle, config)
    return _h


def plugin_poll(h):
    _media_handle = _handle_map.get_handle(h)
    # vector off to the particular medium
    _reading = _media_handle.poll()
    return _reading


def plugin_reconfigure(h, new_config):
    """Reconfigures the plugin

    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _media_handle = _handle_map.get_handle(h)
    _media_handle.reconfigure(new_config)  # may create/destroy rpc server
    h["config"].update(new_config)

    return h


def plugin_shutdown(h):
    """Shutdowns the plugin doing required cleanup, to be called prior to the South plugin service being shut down.

    Args:
        h: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _LOGGER.info("South Edge ML plugin shutdown")
    _media_handle = _handle_map.get_handle(h)
    _media_handle.shutdown()  # shutdown any RPC objects
    _handle_map.del_handle(h)  # remove our mapping
