# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Server functions for plugin module to detect bad bearing """


__author__ = "Deepanshu Yadav"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import json
import os
import sys
import threading
import traceback
import copy
import time
import hashlib
import subprocess
import glob
from datetime import datetime
from contextlib import suppress
import logging
import asyncio
import datetime

import cv2
import numpy as np
import requests
from dateutil import parser

from foglamp.common import logger
from foglamp.common.plugin_helpers import PluginRPCServer
from foglamp.plugins.south.edgeml.web_stream import WebStream
from foglamp.plugins.common import utils
import async_ingest

_LOGGER = logger.setup(__name__, level=logging.INFO)


class FaultyResponse:
    def __init__(self):
        self.status_code = 500


def sendto_edgeml_cluster(frame, url, logger):
    """
        Sends the frame to edgeml cluster and gives the response if successful
        and returns FaultyResponse object if error.
    Args:
        frame (opencv frame): The opencv frame
        url (The url to post): The url of edgeml cluster
        logger (The logger object): The logger object to show error if any.

    Returns:
        The json response from the edgeml cluster
    """
    data = cv2.imencode('.jpg', frame)[1].tobytes()
    files = {'image': data}
    try:
        response = requests.post(url, files=files)
    except Exception as ie:
        logger.error(ie)
        f = FaultyResponse()
        return f

    return response


class Writer(threading.Thread):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.files_list = []
        self.time_delta = datetime.timedelta(minutes=self.parent.rotate_after_minutes)
        self.rotation_time = datetime.datetime.now() + self.time_delta
        self.parent.logger.info("The next time for rotation is {}".format(self.rotation_time))

    def get_no_of_files(self):
        """
            Get the total no of files to rotate based on rotate_data_minutes config parameter.
        Returns:
            The index of self.files_list whose time is just greater than the required time.
            If it is still smaller than returns last index.
        """
        _first_file_name = os.path.split(self.files_list[0])[1]
        _cleaned_first_file_name = _first_file_name.split("_")[1].split(".")[0]
        _required_time = parser.parse(_cleaned_first_file_name) + datetime.\
            timedelta(minutes=self.parent.rotate_data_minutes)

        for index, _file_name in enumerate(self.files_list):
            _file_name = os.path.split(_file_name)[1]
            _cleaned_file_name = _file_name.split("_")[1].split(".")[0]
            _creation_time = parser.parse(_cleaned_file_name)

            if _creation_time >= _required_time:
                return index

        return len(self.files_list) - 1

    def run(self):
        """
        The main loop of writer thread responsible for saving and rotating image files.
        Returns:
            None
        """
        while True:

            _file_list = []
            if self.parent.shutdown_in_progress:
                break

            # This part is mutually exclusive from main thread.
            with self.parent.save_files:
                if self.parent.frame_list is not []:
                    _file_list = self.parent.frame_list
                    self.parent.frame_list = []

                else:
                    # here no files are found to save
                    self.parent.save_files.wait()

            # files are found to save
            # the main thread now will not disturb the saving process.
            for _fname, _frame in _file_list:
                if self.parent.shutdown_in_progress:
                    break
                cv2.imwrite(_fname, _frame)
                self.files_list.append(_fname)

            if not self.parent.shutdown_in_progress:

                # checking for rotation
                if datetime.datetime.now() >= self.rotation_time:
                    self.parent.logger.info("Time for rotation of files")

                    if len(self.files_list) > 0:

                        # getting the number of files to rotate
                        _no_of_files = self.get_no_of_files()
                        for _ in range(_no_of_files):
                            try:
                                _file = self.files_list.pop(0)
                                os.remove(_file)
                                self.parent.logger.info("Removed file {}".format(_file))
                            except IndexError as ie:
                                break

                        # Getting the next time for rotation
                        self.rotation_time = self.rotation_time + self.time_delta
                        self.parent.logger.info("The next time for rotation"
                                                " is {}".format(self.rotation_time))


class Inference(threading.Thread):

    def __init__(self, parent):
        """[]

        Args:
            parent ([type]): [The parent class that controls the this thread
        """
        super().__init__()

        self.parent = parent  # for locks and synchronizatoin
        self.logger = parent.logger
        self.file_list = []
        self.stream = None
        self.initialize_source()

    def initialize_source(self):
        """
        Initializes the video source be it directory, camera or stream.
        Returns:

        """
        if self.parent.source == "directory":

            self.file_list = glob.glob(os.path.join(InferenceServer.FOGLAMP_DATA,
                                                    self.parent.source_dir_name, "*"))
            if not self.file_list:
                self.logger.info("There are no images stored in the mentioned directory.")

        elif self.parent.source == "camera":
            self.stream = cv2.VideoCapture(self.parent.camera_id)
        else:
            os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = "rtsp_transport;udp"
            self.stream = cv2.VideoCapture(self.parent.rtsp_url, cv2.CAP_FFMPEG)

    def get_frame(self):
        """

        Returns:
            The next frame form video source be it directory, camera or stream.
        """
        if self.parent.source == "directory":
            _file_name = None
            try:
                _file_name = self.file_list.pop(0)
            except IndexError as ie:
                # making the files list circular
                self.file_list = glob.glob(os.path.join(InferenceServer.FOGLAMP_DATA,
                                           self.parent.source_dir_name, "*"))
                _file_name = self.file_list.pop(0)

            if _file_name:
                _frame = cv2.imread(_file_name)
                return True, _frame
            else:
                return False, None
        else:
            grabbed, _frame = self.stream.read()
            return grabbed, _frame

    def run(self):
        """[The main thread that keeps on fetching frames and goes to
         sleep for a configurable amount of time.]

        """
        while True:
            start_time = datetime.datetime.now()
            if self.parent.shutdown_in_progress:
                self.logger.info("Thread shutting down")
                WebStream.SHUTDOWN_IN_PROGRESS = True
                break

            grabbed, _frame = self.get_frame()
            frame_arrival_time = datetime.datetime.now()
            fetch_time = (frame_arrival_time - start_time).total_seconds()
            if grabbed:
                _frame = _frame.copy()
                # sending for inference
                _response = sendto_edgeml_cluster(_frame, self.parent.url_to_post, self.logger)
                if _response.status_code == 200:
                    _json_response = _response.json()
                    # drawing the inference
                    self.parent.draw_on_frame(_frame, _json_response)
                    # Saving this frame and generating the readings
                    self.parent.ingest_readings_and_save_frames(_json_response, _frame)

                if self.parent.stream_results:
                    WebStream.FRAME = _frame

            processed_frame_time = datetime.datetime.now()
            processing_time = (processed_frame_time - frame_arrival_time).total_seconds()

            total_time_permitted = 1/(self.parent.fpm/60)
            time_for_sleep = total_time_permitted - (processing_time + fetch_time)
            # the time for sleep is time permitted minus (processing and fetch time)
            if time_for_sleep > 0:
                time.sleep(total_time_permitted)

        # releasing the resource if shutdown
        self.stream.release()


class InferenceServer(PluginRPCServer):
    """[create a queue of images;
    extract images file names when polled and send them upstream as an asset]
    """
    FOGLAMP_ROOT = os.getenv("FOGLAMP_ROOT", default="/usr/local/foglamp")
    FOGLAMP_DATA = os.path.join(FOGLAMP_ROOT, "data")
    EDGEML_REST_RESOURCE = "v1/vision/detection"

    def __init__(self):
        super().__init__("server")

        self.logger = _LOGGER
        self.shutdown_in_progress = False
        self.inference_thread = None
        self.url_to_post = None
        self.loop = None
        self.async_thread = None
        self.web_stream = None
        self.queue = threading.Condition()
        self.readings = []
        self.frame_list = []
        self.save_files = threading.Condition()
        self.writing_thread = None
        self.did_plugin_start = False

        # Fixing the camera height
        self.camera_height = 480
        self.camera_width = 640

        # slots
        self.source = None
        self.fpm = None
        self.source_dir_name = None
        self.camera_id = None
        self.rtsp_url = None
        self.inference_choice = None
        self.deployment_name = None
        self.rest_url = None
        self.stream_results = None
        self.stream_port = None
        self.output_asset = None
        self.destination_dir = None
        self.rotate_after_minutes = None
        self.rotate_data_minutes = None

    def preprocess_config(self):

        # data type handling
        self.source = str(self.source)
        self.fpm = int(self.fpm)
        self.source_dir_name = str(self.source_dir_name)
        self.camera_id = int(self.camera_id)
        self.rtsp_url = str(self.rtsp_url)
        self.inference_choice = str(self.inference_choice)
        self.deployment_name = str(self.deployment_name)
        self.rest_url = str(self.rest_url)
        self.stream_results = bool(self.stream_results)
        self.stream_port = int(self.stream_port)
        self.output_asset = str(self.output_asset)
        self.destination_dir = str(self.destination_dir)
        self.rotate_after_minutes = int(self.rotate_after_minutes)
        self.rotate_data_minutes = int(self.rotate_data_minutes)

    def validate_config(self):
        """
            Make sure the edgeml url works as well video source works.
        Returns:

        """

        if self.source == "directory":
            if not os.path.exists(os.path.join(InferenceServer.FOGLAMP_DATA, self.source_dir_name)):
                self.logger.error("The directory provided does not exist")
                return False

        if self.inference_choice == "k8":
            _p = subprocess.run(["kubectl", "get", "svc", self.deployment_name,
                                 "-o", "jsonpath='{.spec.clusterIP}:{.spec.ports[0].port}'"],
                                stdout=subprocess.PIPE,
                                check=True)
            _url = "http://{}/{}".format(eval(_p.stdout), InferenceServer.EDGEML_REST_RESOURCE)
            self.url_to_post = _url
        else:
            self.url_to_post = self.rest_url
        self.logger.info("The url is {}".format(self.url_to_post))

        # checking the url_to_post
        _test_file_name = os.path.join(os.path.dirname(__file__), "data", "parrot.jpg")
        with open(_test_file_name, 'rb') as fp:
            _files = {'image': fp}
            try:
                r = requests.post(self.url_to_post, files=_files)
            except Exception:
                self.logger.error("The edgeml url is not functional ! ")
                return False

        if r.status_code != 200:
            self.logger.error("The edgeml url is not functional ! ")
            return False

        self.logger.info("The edgeml url is functional ! ")
        # checking the camera/ stream
        if self.source == "camera":
            _stream = cv2.VideoCapture(self.camera_id)
        elif self.source == "stream":
            os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = "rtsp_transport;udp"
            _stream = cv2.VideoCapture(self.rtsp_url, cv2.CAP_FFMPEG)
        else:
            _file_list = glob.glob(os.path.join(InferenceServer.FOGLAMP_DATA,
                                                self.source_dir_name))
            if _file_list:
                self.logger.info(" Files found in given directory")
                return True
            else:
                self.logger.error("No files found in given directory")
                return False

        tries = 1
        max_tries = 3

        while tries < max_tries:
            _grabbed, _frame = _stream.read()
            if _grabbed:
                break
            else:
                time.sleep(1)
                tries += 1

        _stream.release()
        if tries > max_tries:
            self.logger.error("The stream could not start after three attempts")
            return False

        self.logger.info("The config is valid!")
        return True

    def draw_on_frame(self, _frame, _json_dict):
        """
            Draws the detections results on the frame.
        Args:
            _frame (object): The given input frame
            _json_dict (json dict): The detection results

        Returns:
            None
        """

        if "predictions" in _json_dict.keys():
            _detection_array = _json_dict["predictions"]

            if _detection_array:
                for i in range(len(_detection_array)):
                    ymin_model = _detection_array[i]["y_min"]
                    xmin_model = _detection_array[i]["x_min"]
                    ymax_model = _detection_array[i]["y_max"]
                    xmax_model = _detection_array[i]["x_max"]

                    # map the bounding boxes from model to the window
                    ymin = int(max(0, ymin_model))
                    xmin = int(max(0, xmin_model))
                    ymax = int(min(self.camera_height, ymax_model))
                    xmax = int(min(self.camera_width, xmax_model))

                    # draw the rectangle on the frame
                    cv2.rectangle(_frame, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)

                    # Contructing the label

                    # Look up object name from "labels" array using class index
                    object_name = _detection_array[i]["label"]

                    confidence = _detection_array[i]["confidence"]

                    # Example: 'person: 72%'
                    label = '%s: %d%%' % (object_name, int(confidence * 100))

                    # Get font size
                    labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX,
                                                          0.7, 2)

                    # Make sure not to draw label too close to top of window
                    label_ymin = max(ymin, labelSize[1] + 10)

                    # Draw white box to put label text in
                    cv2.rectangle(_frame, (xmin, label_ymin - labelSize[1] - 10),
                                  (xmin + labelSize[0], label_ymin + baseLine - 10),
                                  (255, 255, 255), cv2.FILLED)

                    # Draw the text label
                    cv2.putText(_frame, label, (xmin, label_ymin - 7),
                                cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2)

    def save_file(self, _file_name, _frame):
        """

        Args:
            _file_name (str): The file name associated with the frame.
            _frame (opencv frame): The frame to be saved as jpeg file.

        Returns:
            None
        """

        # this part is mutually exclusive. so that if just in case writer thread is taking
        # files to save at the same time.
        with self.save_files:
            # append to frame list
            self.frame_list.append((_file_name, _frame))
            # dont save it just notify the writer thread.
            self.save_files.notify()

    def ingest_readings_and_save_frames(self, _json_response, _frame):
        """
            Generates readings from json response
        Args:
            _json_response (json string): The response from edgeml cluster
            _frame (opencv frame) : the opencv frame
        Returns:
            readings (dict) The readings that are going to be ingested into database.
        """
        if "predictions" in _json_response.keys():
            reads = _json_response["predictions"]
        else:
            reads = []

        _time_stamp = utils.local_timestamp()
        _file_name = "IMG" + "_"+str(_time_stamp) + ".jpg"
        _full_file_name = os.path.join(InferenceServer.FOGLAMP_DATA, self.destination_dir, _file_name)
        self.save_file(_full_file_name, _frame)
        reads.append({"file_name": _full_file_name})
        data = {
            'asset': self.output_asset,
            'timestamp': _time_stamp,
            'readings': reads
        }

        with self.queue:
            self.readings.append(data)

            # just in case reading get clogged over time.
            if len(self.readings) > 100:
                self.logger.debug("PRODUCE: overflow")
                self.readings.pop(0)

            # inform the poll function that readings are available.
            self.queue.notify()  # signal to processing thread

    def start_threads(self):
        """
        Starts all the required threads
        Returns:

        """
        self.inference_thread = Inference(self)
        self.inference_thread.start()

        self.writing_thread = Writer(self)
        self.writing_thread.start()

        if self.stream_results:
            self.loop = asyncio.new_event_loop()
            # make shutdown flag of web stream server false.
            WebStream.SHUTDOWN_IN_PROGRESS = False

            self.web_stream = WebStream(port=self.stream_port).start_web_streaming_server(
                local_loop=self.loop)
            self.async_thread = threading.Thread(target=self.run_async, name="Async Thread")
            self.async_thread.daemon = True
            self.async_thread.start()

    # ### PLUGIN API ### #
    def plugin_init(self, config):
        """[Plugin init at the server side.]

        Args:
            config ([dict]): [A configuration from client side.]

        Returns:
            [dict]: [returns the same config as it is.]
        """
        # called to initialize or reinitialize camera
        self.logger.info("PLUGIN INIT INFERENCE SERVER")
        self.config_update(config)   # transfer config elements to handle
        self.preprocess_config()
        is_config_valid = self.validate_config()

        if is_config_valid:
            # it is safe to start the plugin
            _dest_dir = os.path.join(InferenceServer.FOGLAMP_DATA, self.destination_dir)
            if not os.path.exists(_dest_dir):
                os.makedirs(_dest_dir)
            self.start_threads()
            self.did_plugin_start = True
        else:
            self.logger.error("Please shut the plugin down and try again")

    def run_async(self):
        self.loop.run_forever()

    def plugin_poll(self):
        """

        Returns: Returns readings for poll

        """
        _readings = []
        if self.did_plugin_start:

            # make sure we are not in conflict with readings generation.
            with self.queue:
                while len(_readings) == 0:
                    if len(self.readings) == 0:
                        # wait if no readings available
                        self.queue.wait()
                    else:
                        # return the readings if found.
                        _readings = self.readings
                        self.readings = []
        else:
            self.logger.info("Plugin did not started will return empty readings.")
        return _readings

    def plugin_shutdown(self):
        """[shutdown at the server side.]
        """
        self.logger.info("Server shutting down")
        self.shutdown_in_progress = True
        if self.inference_thread is not None:
            self.inference_thread.join()
        if self.writing_thread is not None:
            self.writing_thread.join()
        if self.stream_results:
            if self.web_stream is not None:
                WebStream.SHUTDOWN_IN_PROGRESS = True
                self.web_stream.stop_server(self.loop)

            if self.loop is not None:
                self.loop.stop()
            self.async_thread = None
            self.loop = None

        self.writing_thread = None
        self.inference_thread = None

    def plugin_reconfigure(self, members_dict):
        """
        Plugin reconfigure at server side.
        Args:
            members_dict (dict): The dictionary which will change variables.

        Returns:
            None
        """
        self.logger.info("Plugin reconfigure inside server called.")
        for key, value in members_dict.items():
            setattr(self, key, value)

        _dest_dir = os.path.join(InferenceServer.FOGLAMP_DATA, self.destination_dir)
        if not os.path.exists(_dest_dir):
            os.makedirs(_dest_dir)
        self.writing_thread.time_delta = datetime.timedelta(minutes=self.rotate_after_minutes)


_LOGGER.info("Inference server SERVING....")
rpc = InferenceServer()
rpc.serve()
